{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "a75ede6d-7c9d-48fa-bdb6-b55f0d9ffcc2",
   "metadata": {},
   "source": [
    "# Introduction to OpenMp (Shared memory)\n",
    "\n",
    "This is an extension for languages like C/C++. It is targeted for\n",
    "**multi-threaded, shared memory** parallelism. It is very simple, you\n",
    "just add some pragma directives which, if not supported, are deactivated\n",
    "and then completely ignored. It is ideal for a machine with several\n",
    "cores, and shared memory. For much more info, see\n",
    "<http://www.openmp.org/> and <https://hpc-tutorials.llnl.gov/openmp/>\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "7aaf957e",
   "metadata": {},
   "source": [
    "## Introduction: Hello world and threadid\n",
    "\n",
    "Typical Hello world,\n",
    "\n",
    "``` cpp\n",
    "#include <cstdio>\n",
    "#include \"omp.h\"\n",
    "\n",
    "int main(void)\n",
    "{\n",
    "    double x = 9.0;\n",
    "//#pragma omp parallel num_threads(4)\n",
    "#pragma omp parallel\n",
    "    {// se generan los threads\n",
    "        std::printf(\"Hello, world.\\n\");\n",
    "        std::printf(\"Hello, world2.\\n\");\n",
    "    } // mueren los threads\n",
    "\n",
    "    return 0;\n",
    "}\n",
    "```\n",
    "\n",
    "This makes a parallel version if you compile with the `-fopemp` flag.\n",
    "Test it. What happens if you write\n",
    "\n",
    "``` bash\n",
    "export OMP_NUM_THREADS=8\n",
    "```\n",
    "\n",
    "and then you run the executable?\n",
    "\n",
    "The actual directive is like\n",
    "\n",
    "``` example\n",
    "#pragma omp parallel [clause ...]  newline \n",
    "                   if (scalar_expression) \n",
    "                   private (list) \n",
    "                   shared (list) \n",
    "                   default (shared | none) \n",
    "                   firstprivate (list) \n",
    "                   reduction (operator: list) \n",
    "                   copyin (list) \n",
    "                   num_threads (integer-expression)\n",
    " structured_block  \n",
    "```\n",
    "\n",
    "But we will keep things simple.\n",
    "\n",
    "What is really important to keep in mind is what variables are going to\n",
    "be shared and what are going to be private, to avoid errors, reca\n",
    "conditions, etc.\n",
    "\n",
    "If you want to know the thread id, you can use\n",
    "\n",
    "``` c\n",
    "int nth =  omp_get_num_threads();\n",
    "int tid =  omp_get_thread_num();\n",
    "```\n",
    "\n",
    "inside your code, as in the following,\n",
    "\n",
    "``` cpp\n",
    "#include <cstdio> // printf\n",
    "#include <omp.h>\n",
    "\n",
    "int main(void)\n",
    "{\n",
    "  double x = 9.0;\n",
    "\n",
    "  int nth = omp_get_num_threads();\n",
    "  int thid = omp_get_thread_num();\n",
    "  std::printf(\"Hello world from thid: %d, out of %d .\\n\",\n",
    "              thid, nth);\n",
    "\n",
    "//#pragma omp parallel num_threads(4)\n",
    "#pragma omp parallel\n",
    "  {// se generan los threads\n",
    "    int nth = omp_get_num_threads(); // al declarar aca, son privados\n",
    "    int thid = omp_get_thread_num();\n",
    "    std::printf(\"Hello world from thid: %d, out of %d .\\n\",\n",
    "                thid, nth);\n",
    "  } // mueren los threads\n",
    "\n",
    "  std::printf(\"Hello world from thid: %d, out of %d .\\n\",\n",
    "              thid, nth);\n",
    "\n",
    "\n",
    "  return 0;\n",
    "}\n",
    "```\n",
    "\n",
    "Please use the thread sanitize:\n",
    "\n",
    "``` shell\n",
    "g++ -fopenmp -g -fsanitize=thread code.cpp\n",
    "```\n",
    "\n",
    "There are some other env variables that could be useful:\n",
    "\n",
    "``` shell\n",
    "OMP_NUM_THREADS=4\n",
    "OMP_DISPLAY_ENV=TRUE\n",
    "OMP_DISPLAY_AFFINITY=TRUE\n",
    "OMP_STACK_SIZE=1000000\n",
    "```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "c4d029a3",
   "metadata": {},
   "source": [
    "\n",
    "## Private and shared variables\n",
    "\n",
    "Shared and private\n",
    "\n",
    "``` cpp\n",
    "#include <cstdio>\n",
    "#include <omp.h>\n",
    "\n",
    "int main(void)\n",
    "{\n",
    "  double x = 9.0;\n",
    "\n",
    "  int nth = omp_get_num_threads();\n",
    "  int thid = omp_get_thread_num();\n",
    "  std::printf(\"Hello world from thid: %d, out of %d .\\n\",\n",
    "              thid, nth);\n",
    "\n",
    "//#pragma omp parallel num_threads(4)\n",
    "#pragma omp parallel private(thid, nth)\n",
    "  {// se generan los threads\n",
    "    thid = omp_get_thread_num(); // privada, aunque tenga el mismo nombre, tipo, y no haya sido redeclarada\n",
    "    nth = omp_get_num_threads(); // privada, aunque tenga el mismo nombre, tipo, y no haya sido redeclarada\n",
    "    std::printf(\"Hello world from thid: %d, out of %d .\\n\",\n",
    "                thid, nth);\n",
    "  } // mueren los threads\n",
    "\n",
    "  return 0;\n",
    "}\n",
    "```\n",
    "In this example we see the memory address for shared and private\n",
    "variables to illustrate those concepts:\n",
    "\n",
    "``` cpp\n",
    "#include <omp.h>\n",
    "#include <iostream>\n",
    "#include <cstdio>\n",
    "\n",
    "int main(int argc, char *argv[]) {\n",
    "\n",
    "  int nthreads, tid;\n",
    "\n",
    "  /* Fork a team of threads with each thread having a private tid variable */\n",
    "#pragma omp parallel private(tid)\n",
    "  {\n",
    "\n",
    "    /* Obtain and print thread id */\n",
    "    tid = omp_get_thread_num();\n",
    "    std::printf(\"Hello World from thread = %d\\n\", tid);\n",
    "    std::cout << \"Memory address for tid = \" << &tid << std::endl;\n",
    "    std::cout << \"Memory address for nthreads = \" << &nthreads << std::endl;\n",
    "\n",
    "    /* Only master thread does this */\n",
    "    if (tid == 0) \n",
    "      {\n",
    "        nthreads = omp_get_num_threads();\n",
    "        printf(\"Number of threads = %d\\n\", nthreads);\n",
    "      }\n",
    "\n",
    "  }  /* All threads join master thread and terminate */\n",
    "\n",
    "  return 0;\n",
    "}\n",
    "```\n",
    "\n",
    "\n",
    "Selecting code only for the master thread\n",
    "``` c\n",
    "#include <omp.h>\n",
    "#include <stdio.h>\n",
    "\n",
    "int main(int argc, char *argv[]) {\n",
    "\n",
    "  int nthreads, tid;\n",
    "\n",
    "  /* Fork a team of threads with each thread having a private tid variable */\n",
    "#pragma omp parallel private(tid)\n",
    "  {\n",
    "\n",
    "    /* Obtain and print thread id */\n",
    "    tid = omp_get_thread_num();\n",
    "    printf(\"Hello World from thread = %d\\n\", tid);\n",
    "\n",
    "    /* Only master thread does this */\n",
    "    if (tid == 0)\n",
    "    {\n",
    "      nthreads = omp_get_num_threads();\n",
    "      printf(\"Number of threads = %d\\n\", nthreads);\n",
    "    }\n",
    "\n",
    "  }  /* All threads join master thread and terminate */\n",
    "\n",
    "}\n",
    "```"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "0c7bcb42",
   "metadata": {},
   "source": [
    "\n",
    "## Parallelizing a for loop\n",
    "\n",
    "Condiciones del problema:\n",
    "\n",
    "-   Arreglo de tamanho N.\n",
    "-   Tenemos nth threads.\n",
    "-   `Nlocal = N/nth`\n",
    "-   `imin = tid*Nlocal, imax = (tid+1)*Nlocal = imin + Nlocal`\n",
    "\n",
    "``` cpp\n",
    "\n",
    "```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "9ba00251",
   "metadata": {},
   "source": [
    "\n",
    "### Manually\n",
    "\n",
    "The following code assigns the elements of a vector in parallel:\n",
    "\n",
    "``` c\n",
    "#include <omp.h>\n",
    "#include <iostream>\n",
    "#include <cmath>\n",
    "#include <vector>\n",
    "#include <cstdlib>\n",
    "#include <numeric>\n",
    "\n",
    "void fill(std::vector<double> & array);\n",
    "double suma(const std::vector<double> & array);\n",
    "\n",
    "int main(int argc, char *argv[])\n",
    "{\n",
    "  const int N = std::atoi(argv[1]);\n",
    "  std::vector<double> data(N);\n",
    "\n",
    "  // llenar el arreglo\n",
    "  fill(data);\n",
    "  //std::cout << data[0] << \"\\n\";\n",
    "\n",
    "  // calcular la suma y el promedio\n",
    "  double total = suma(data);\n",
    "  std::cout << total/data.size() << \"\\n\";\n",
    "\n",
    "  return 0;\n",
    "}\n",
    "\n",
    "void fill(std::vector<double> & array)\n",
    "{\n",
    "  int N = array.size();\n",
    "#pragma omp parallel\n",
    "  {\n",
    "    int thid = omp_get_thread_num();\n",
    "    int nth = omp_get_num_threads();\n",
    "    int Nlocal = N/nth;\n",
    "    int iimin = thid*Nlocal;\n",
    "    int iimax = iimin + Nlocal;\n",
    "    for(int ii = iimin; ii < iimax; ii++) {\n",
    "      array[ii] = 2*ii*std::sin(std::sqrt(ii/56.7)) +\n",
    "        std::cos(std::pow(1.0*ii*ii/N, 0.3));\n",
    "    }\n",
    "  }\n",
    "}\n",
    "\n",
    "double suma(const std::vector<double> & array)\n",
    "{\n",
    "  int N = array.size();\n",
    "  std::vector<double> sumaparcial;\n",
    "#pragma omp parallel\n",
    "  {\n",
    "    int thid = omp_get_thread_num();\n",
    "    int nth = omp_get_num_threads();\n",
    "    if (0 == thid) {\n",
    "      sumaparcial.resize(nth);\n",
    "    }\n",
    "#pragma omp barrier\n",
    "    double localsum = 0.0;\n",
    "    int Nlocal = N/nth;\n",
    "    int iimin = thid*Nlocal;\n",
    "    int iimax = iimin + Nlocal;\n",
    "    for(int ii = iimin; ii < iimax; ii++) {\n",
    "      localsum += array[ii];\n",
    "    }\n",
    "    sumaparcial[thid] = localsum;\n",
    "  }\n",
    "  return std::accumulate(sumaparcial.begin(), sumaparcial.end(), 0.0);\n",
    "}\n",
    "\n",
    "\n",
    "```\n",
    "\n",
    "OLD\n",
    "\n",
    "``` c\n",
    "#include <omp.h>\n",
    "#include <iostream>\n",
    "#include <cmath>\n",
    "#include <vector>\n",
    "\n",
    "int main(int argc, char *argv[]) {\n",
    "\n",
    "  const int N = 80000000;\n",
    "  int i;\n",
    "  double *a = new double[N];\n",
    "\n",
    "#pragma omp parallel private(i)\n",
    "  {\n",
    "    int thid = omp_get_thread_num();\n",
    "    int nth = omp_get_num_threads();\n",
    "    int localsize = N/nth;\n",
    "    int iimin = thid*localsize;\n",
    "    int iimax = iimin + localsize;\n",
    "    for(i = iimin; i < iimax; i++) {\n",
    "      a[i] = 2*i*std::sin(std::sqrt(i/56.7)) +\n",
    "        std::cos(std::pow(i*i, 0.3));\n",
    "    }\n",
    "  }\n",
    "\n",
    "  std::cout << a[1] << \"\\n\";\n",
    "\n",
    "  delete [] a;\n",
    "  return 0;\n",
    "}\n",
    "```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "e6dee8af",
   "metadata": {},
   "source": [
    "\n",
    "### Using `omp parallel for`\n",
    "\n",
    "``` c\n",
    "#include <omp.h>\n",
    "#include <iostream>\n",
    "#include <cmath>\n",
    "#include <vector>\n",
    "#include <cstdlib>\n",
    "#include <numeric>\n",
    "\n",
    "void fill(std::vector<double> & array);\n",
    "double suma(const std::vector<double> & array);\n",
    "\n",
    "int main(int argc, char *argv[])\n",
    "{\n",
    "  const int N = std::atoi(argv[1]);\n",
    "  std::vector<double> data(N);\n",
    "\n",
    "  // llenar el arreglo\n",
    "  fill(data);\n",
    "  //std::cout << data[0] << \"\\n\";\n",
    "\n",
    "  // calcular la suma y el promedio\n",
    "  double total = suma(data);\n",
    "  std::cout << total/data.size() << \"\\n\";\n",
    "\n",
    "  return 0;\n",
    "}\n",
    "\n",
    "void fill(std::vector<double> & array)\n",
    "{\n",
    "  const int N = array.size();\n",
    "#pragma omp parallel for\n",
    "  for(int ii = 0; ii < N; ii++) {\n",
    "      array[ii] = 2*ii*std::sin(std::sqrt(ii/56.7)) +\n",
    "        std::cos(std::pow(1.0*ii*ii/N, 0.3));\n",
    "  }\n",
    "}\n",
    "\n",
    "double suma(const std::vector<double> & array)\n",
    "{\n",
    "  int N = array.size();\n",
    "  double suma = 0.0;\n",
    "#pragma omp parallel for reduction(+:suma)\n",
    "  for(int ii = 0; ii < N; ii++) {\n",
    "    suma += array[ii];\n",
    "  }\n",
    "  return suma;\n",
    "}\n",
    "\n",
    "\n",
    "```\n",
    "\n",
    "OLD\n",
    "\n",
    "``` c\n",
    "#include <omp.h>\n",
    "#include <iostream>\n",
    "#include <cmath>\n",
    "\n",
    "void fill(double * array, int size);\n",
    "double average(double * array, int size);\n",
    "\n",
    "int main(int argc, char *argv[]) {\n",
    "  const int N = std::atoi(argv[1]);\n",
    "  double *a = new double[N];\n",
    "\n",
    "  fill(a, N);\n",
    "  std::cout << a[1] << \"\\n\";\n",
    "\n",
    "  //std::cout << average(a, N) << \"\\n\";\n",
    "\n",
    "  double suma = 0.0;\n",
    "#pragma omp parallel for reduction(+:suma)\n",
    "  for(int i = 0; i < N; i++) {\n",
    "    suma += a[i];\n",
    "  }\n",
    "  std::cout <<  suma/N << \"\\n\";\n",
    "\n",
    "  delete [] a;\n",
    "  return 0;\n",
    "}\n",
    "\n",
    "void fill(double * array, int size)\n",
    "{\n",
    "#pragma omp parallel for\n",
    "  for(int i = 0; i < size; i++) {\n",
    "    array[i] = 2*i*std::sin(std::sqrt(i/56.7)) +\n",
    "      std::cos(std::pow(1.0*i*i, 0.3));\n",
    "  }\n",
    "}\n",
    "double average(double * array, int size)\n",
    "{\n",
    "  double suma = 0.0;\n",
    "#pragma omp parallel for reduction(+:suma)\n",
    "  for(int i = 0; i < size; i++) {\n",
    "    suma += array[i];\n",
    "  }\n",
    "  return suma/size;\n",
    "}\n",
    "\n",
    "```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "620b5b07",
   "metadata": {},
   "source": [
    "\n",
    "**Exercise:** Modify the code to be able to compute the mean in\n",
    "parallel. Play with larger and larger arrays and also with increasing\n",
    "the number of processors. Can you compute any parallel metric?\n",
    "\n",
    "``` cpp\n",
    "#include <omp.h>\n",
    "#include <iostream>\n",
    "#include <cmath>\n",
    "\n",
    "void fill_data(double *a, int size);\n",
    "double average(double *a, int size);\n",
    "\n",
    "int main(int argc, char *argv[]) {\n",
    "    std::cout.precision(15); std::cout.setf(std::ios::scientific);\n",
    "\n",
    "    const int N = 80000000;\n",
    "    int i;\n",
    "    double *a = new double[N];\n",
    "\n",
    "    fill_data(a, N);\n",
    "    std::cout << a[1] << std::endl;\n",
    "    double avg = average(a, N);\n",
    "    std::cout << avg << \"\\n\";\n",
    "\n",
    "    delete [] a;\n",
    "    return 0;\n",
    "}\n",
    "\n",
    "void fill_data(double *a, int size)\n",
    "{\n",
    "    long int i;\n",
    "#pragma omp parallel for\n",
    "    for(i = 0; i < size; i++) {\n",
    "        a[i] = 2*i*std::sin(std::sqrt(i/56.7)) +\n",
    "            std::cos(std::pow(i*i, 0.3));\n",
    "    }\n",
    "}\n",
    "\n",
    "double average(double *a, int size)\n",
    "{\n",
    "    double result = 0;\n",
    "#pragma omp parallel for reduction(+:result)\n",
    "    for(int ii = 0; ii < size; ++ii) {\n",
    "      result = result + a[ii];\n",
    "    }\n",
    "    result /= size;\n",
    "    return result;\n",
    "}\n",
    "```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "1d9dfaf0",
   "metadata": {},
   "source": [
    "\n",
    "### Using reductions\n",
    "\n",
    "This example shows how to reduce a variable using the `reduction`\n",
    "keyword:\n",
    "\n",
    "``` cpp\n",
    "  #include <iostream>\n",
    "#include <cmath>\n",
    "\n",
    "int main(int argc, char ** argv)\n",
    "{\n",
    "  const int N = 100000000;\n",
    "  int i;\n",
    "  double *a;\n",
    "  a = new double [N];\n",
    "  double suma = 0;\n",
    "\n",
    "#pragma omp parallel for reduction(+:suma)\n",
    "  for (i = 0; i < N; ++i) {\n",
    "    a[i] = 2*std::sin(i/35.6);\n",
    "    suma += a[i];\n",
    "  }\n",
    "\n",
    "  std::cout << suma << std::endl;\n",
    "\n",
    "  delete [] a;\n",
    "  return 0;\n",
    "}\n",
    "```\n",
    "\n",
    "Now read the following code, try to predict its outcome, and then run it\n",
    "and test your prediction (from\n",
    "<https://computing.llnl.gov/tutorials/openMP/#Introduction>) :\n",
    "\n",
    "``` c\n",
    "#include <omp.h>\n",
    "#define N 1000\n",
    "#define CHUNKSIZE 100\n",
    "\n",
    "main(int argc, char *argv[]) {\n",
    "\n",
    "  int i, chunk;\n",
    "  float a[N], b[N], c[N];\n",
    "\n",
    "  /* Some initializations */\n",
    "  for (i=0; i < N; i++)\n",
    "    a[i] = b[i] = i * 1.0;\n",
    "  chunk = CHUNKSIZE;\n",
    "\n",
    "#pragma omp parallel shared(a,b,c,chunk) private(i)\n",
    "  {\n",
    "\n",
    "#pragma omp for schedule(dynamic,chunk) nowait\n",
    "    for (i=0; i < N; i++)\n",
    "c[i] = a[i] + b[i];\n",
    "\n",
    "  }   /* end of parallel region */\n",
    "  return 0;\n",
    "}\n",
    "```\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "12b1869f",
   "metadata": {},
   "source": [
    "## Tutorial and extra materials:\n",
    "-   <https://researchcomputing.princeton.edu/education/external-online-resources/openmp>\n",
    "-   <https://cvw.cac.cornell.edu/openmp/default>\n",
    "-   Scheduling:\n",
    "    -   <https://610yilingliu.github.io/2020/07/15/ScheduleinOpenMP/>\n",
    "    -   <https://www.openmp.org/wp-content/uploads/SC17-Kale-LoopSchedforOMP_BoothTalk.pdf>\n",
    "    -   <http://jakascorner.com/blog/2016/06/omp-for-scheduling.html>\n",
    "\n",
    "Parallel STL\n",
    "<https://www.modernescpp.com/index.php?option=com_content&view=article&id=572&catid=49>\n",
    "<https://www.modernescpp.com/index.php?option=com_content&view=article&id=573&catid=49>\n"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "7f369569",
   "metadata": {},
   "source": [
    "\n",
    "## Exercises\n",
    "\n",
    "Refs: CT-LAB openmp exercises;\n",
    "<https://computing.llnl.gov/tutorials/openMP/exercise.html>\n",
    "\n",
    "1.  Write a program to compute the mean and the standard deviation for a\n",
    "    large array using OpenMP, with only one reduction. Also compute the\n",
    "    parallel metrics.\n",
    "\n",
    "    ``` cpp\n",
    "    // stats.cpp\n",
    "    #include <omp.h>\n",
    "    #include <iostream>\n",
    "    #include <cmath>\n",
    "    #include <vector>\n",
    "    #include <cstdlib>\n",
    "    #include <numeric>\n",
    "\n",
    "    void fill(std::vector<double> & array);\n",
    "    void stats(const std::vector<double> & array, double &mean, double &sigma);\n",
    "\n",
    "    int main(int argc, char *argv[])\n",
    "    {\n",
    "      const int N = std::atoi(argv[1]);\n",
    "      std::vector<double> data(N);\n",
    "\n",
    "      // llenar el arreglo\n",
    "      fill(data);\n",
    "\n",
    "      // calcular stats\n",
    "      double mean{0.0}, sigma{0.0};\n",
    "      double start = omp_get_wtime();\n",
    "      stats(data, mean, sigma);\n",
    "      double time = omp_get_wtime() - start;\n",
    "      std::printf(\"%.15le\\t\\t%.15le\\t\\t%.15le\\n\", mean, sigma, time);\n",
    "\n",
    "      return 0;\n",
    "    }\n",
    "\n",
    "    void fill(std::vector<double> & array)\n",
    "    {\n",
    "      const int N = array.size();\n",
    "    #pragma omp parallel for\n",
    "      for(int ii = 0; ii < N; ii++) {\n",
    "          array[ii] = 2*ii*std::sin(std::sqrt(ii/56.7)) +\n",
    "            std::cos(std::pow(1.0*ii*ii/N, 0.3));\n",
    "      }\n",
    "    }\n",
    "\n",
    "    void stats(const std::vector<double> & array, double & mean, double & sigma)\n",
    "    {\n",
    "      int N = array.size();\n",
    "      double suma = 0.0;\n",
    "    #pragma omp parallel for reduction(+:suma)\n",
    "      for(int ii = 0; ii < N; ii++) {\n",
    "        suma += array[ii];\n",
    "      }\n",
    "      mean = suma/N;\n",
    "    }\n",
    "\n",
    "    ```\n",
    "\n",
    "2.  Write a program to compute the integral of the function $y =\n",
    "      x^{2}$, for \\$ x  ∈ \\[0,10\\]\\$. Fill the following table:\n",
    "\n",
    "    | \\# Threads | Runtime \\[s\\] | Speedup | Efficiency |\n",
    "    |------------|---------------|---------|------------|\n",
    "\n",
    "    Explain your results in terms of the number of processors available.\n",
    "    Use $N=12000$. Do it in two ways:\n",
    "\n",
    "    1.  Distributing the N intervals across all threads, so each one has\n",
    "        a smaller part.\n",
    "    2.  Keeping constant the resolution per thread, that is , each\n",
    "        trhead has N intervals.\n",
    "\n",
    "    What is the difference between the two cases, regarding the\n",
    "    precision and the time?\n",
    "\n",
    "``` cpp\n",
    "#include <iostream>\n",
    "#include <omp.h>\n",
    "\n",
    "using fptr = double(double);\n",
    "\n",
    "double f(double x);\n",
    "double integral_serial(double a, double b, int N, fptr f);\n",
    "double integral_openmp(double a, double b, int N, fptr f);\n",
    "\n",
    "int main(void)\n",
    "{\n",
    "  // declare vars\n",
    "  const double XA = 0.0; \n",
    "  const double XB = 10.0; \n",
    "  const int N = 100000000;\n",
    "\n",
    "  // print result\n",
    "  //std::cout << \"Serial integral: \" << integral_serial(XA, XB, N, f) << \"\\n\";\n",
    "  //std::cout << \"Serial openmp  : \" << integral_openmp(XA, XB, N, f) << \"\\n\";\n",
    "  double t1 = omp_get_wtime();\n",
    "  integral_openmp(XA, XB, N, f);\n",
    "  double t2 = omp_get_wtime();\n",
    "\n",
    "#pragma omp parallel\n",
    "  {\n",
    "    if(0 == omp_get_thread_num()) {\n",
    "      std::cout << omp_get_num_threads() << \"\\t\" << t2 - t1 << std::endl;\n",
    "    }\n",
    "  }\n",
    "}\n",
    "\n",
    "double f(double x)\n",
    "{\n",
    "  return x*x;\n",
    "}\n",
    "\n",
    "double integral_serial(double a, double b, int N, fptr f)\n",
    "{\n",
    "  const double dx = (b-a)/N; \n",
    "  // compute integral\n",
    "  double sum = 0, x;\n",
    "  for(int ii = 0; ii < N; ++ii) {\n",
    "    x = a + ii*dx;\n",
    "    sum += dx*f(x);\n",
    "  }\n",
    "  return sum;\n",
    "}\n",
    "\n",
    "double integral_openmp(double a, double b, int N, fptr f)\n",
    "{\n",
    "  TODO\n",
    "}\n",
    "\n",
    "```\n",
    "\n",
    "1.  Parallelize a Matrix-Matrix multiplication. Compare the performance\n",
    "    when you use one, two, three, for threads.\n",
    "\n",
    "2.  Is this loop parallelizable? If not, why?\n",
    "\n",
    "    ``` cpp\n",
    "    #pragma omp parallel for\n",
    "    for (int i = 1; i < N; i++)\n",
    "    {\n",
    "      A[i] = B[i] – A[i – 1];\n",
    "    }\n",
    "    ```\n",
    "\n",
    "3.  Parallelize a matrix-vector multiplication. What must be shared?\n",
    "    what should be private?\n",
    "\n",
    "4.  Parallelize the matrix transposition.\n",
    "\n",
    "5.  Check and solve the exercises on\n",
    "    <https://computing.llnl.gov/tutorials/openMP/exercise.html> .\n",
    "\n",
    "6.  Check and solve <http://www.hpc.cineca.it/content/training-openmp> .\n",
    "\n",
    "7.  Some other examples are at\n",
    "    <https://www.archer.ac.uk/training/course-material/2015/03/thread_prog/>\n",
    "    ."
   ]
  }
 ],
 "metadata": {
  "language_info": {
   "name": "python"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
